#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <linux/soundcard.h>

char *progname;
int mix_fd = -1;

#define NDEVS SOUND_MIXER_NRDEVICES

char *labels[] = SOUND_DEVICE_LABELS;
char *names[] = SOUND_DEVICE_NAMES;
int devs, recable, stereo, caps;

void die(char *syscall)
{
  perror(syscall);
  exit(1);
}

void Open()
{
  if((mix_fd = open("/dev/mixer",O_RDWR)) <0) die("open");
  if(ioctl(mix_fd, SOUND_MIXER_READ_DEVMASK, &devs)==-1) die("SOUND_MIXER_READ_DEVMASK");
  if(ioctl(mix_fd, SOUND_MIXER_READ_RECMASK, &recable)==-1) die("SOUND_MIXER_READ_RECMASK");
  if(ioctl(mix_fd, SOUND_MIXER_READ_STEREODEVS, &stereo)==-1) die("SOUND_MIXER_READ_STEREODEVS");
  if(ioctl(mix_fd, SOUND_MIXER_READ_CAPS, &caps)==-1) die("SOUND_MIXER_READ_CAPS");
}

void Close()
{
  close(mix_fd);
  mix_fd=-1;
}

int GetRecSrc()
{
  int mask;
  if(ioctl(mix_fd, SOUND_MIXER_READ_RECSRC, &mask)==-1) die("SOUND_MIXER_READ_RECSRC");
  return mask;
}

void SetRecSrc(int mask)
{
  if(ioctl(mix_fd, SOUND_MIXER_WRITE_RECSRC, &mask)==-1) die("SOUND_MIXER_READ_RECSRC");
}

void Get(int c, int *l, int *r)
{
  int v;
  if(ioctl(mix_fd, MIXER_READ(c), &v)==-1) die("MIXER_READ");
  *l = v & 0xFF;
  *r = (v >> 8) & 0xFF;
}

void Set(int c, int l, int r)
{
  int v;
  if(!(stereo & (1<<c)) && l!=r)
    { fprintf(stderr,"Unbalanced setting for mono channel %s\n",
	      names[c]);
      exit(1);
    }
  v = ((r & 0xFF)<<8) + (l & 0xFF);
  if(ioctl(mix_fd, MIXER_WRITE(c), &v)==-1) die("MIXER_WRITE");
}

void PrintDev(int dev)
{
  int left, right;
  if(recable & (1<<dev)) printf("R ");
  else printf("  ");
  printf("%s:  ", labels[dev]);
  Get(dev,&left,&right);
  if((stereo & (1<<dev)) && left!=right)
    printf("%d,%d", left,right);
  else printf("%d", left);
  if(!(stereo & (1<<dev)))
    printf(" (mono)");
  printf("\n");
}

void DumpPrintDev(int dev, FILE *fid)
{
  int left, right;
  fprintf(fid,"%s ", labels[dev]);
  Get(dev,&left,&right);
  if((stereo & (1<<dev)) && left!=right)
    fprintf(fid,"%d,%d", left,right);
  else fprintf(fid,"%d", left);
  fprintf(fid,"\n");
}

void PrintRecSrc()
{
  int i, mask;
  printf("Current recording sources:");
  mask = GetRecSrc();
  for(i=0; i<NDEVS; i++)
    if(mask & (1<<i))
      printf(" %s", names[i]);
  printf("\n");
}

void DumpPrintRecSrc(FILE *fid)
{
  int i, mask;
  fprintf(fid,"recsrc");
  mask = GetRecSrc();
  for(i=0; i<NDEVS; i++)
    if(mask & (1<<i))
      fprintf(fid," %s", names[i]);
  fprintf(fid,"\n");
}

void ListAll()
{
  int i;
  for(i=0; i<NDEVS; i++)
    if(devs & (1<<i))
      PrintDev(i);
  printf("\n");
  if(caps & SOUND_CAP_EXCL_INPUT)
    printf("Only one recording source at a time.\n");
  else printf("Multiple simultaneous recording sources allowed.\n");
  PrintRecSrc();
}

void DumpFile(char *name)
{
  int i;
  FILE *fid;
  if(!strcmp(name,"-")) fid = stdout;
  else if((fid = fopen(name,"w")) == 0) die("fopen");
  for(i=0; i<NDEVS; i++)
    if(devs & (1<<i))
      DumpPrintDev(i,fid);
  fprintf(fid,"\n");
  DumpPrintRecSrc(fid);
  if(strcmp(name,"-")) fclose(fid);
}

int WhichDev(char *name)
{
  int i, dev=-1, len, l1,l2;
  len = strlen(name);
  for(i=0; i<NDEVS; i++){
    if(devs & (1<<i)){
      l1 = strlen(labels[i]);
      l2 = strlen(names[i]);
      if((len<=l1 && !strncmp(name,labels[i],len))
	 || (len<=l2 && !strncmp(name,names[i],len))){
	if(len==l1 || len==l2) return(i);
	else if(dev == -1) dev = i;
      }
    }
  }
  if(dev==-1) { fprintf(stderr,"Channel %s not recognized.\n", name); exit(1); }
  return(dev);
}

int GetPair(char *dat, int *left, int *right)
{
  char *endptr;
  int n;
  n = strtol(dat, &endptr, 10);
  if(endptr == dat || n<0 || n>100) return(0);
  if(*endptr == 0){
    *left = *right = n;
    return(1);
  }
  if(*endptr != ',') return(0);
  *left = n;
  dat = endptr+1;
  n = strtol(dat, &endptr, 10);
  if(*endptr != 0 || n<0 || n>100) return(0);
  *right = n;
  return(1);
}  

void Usage()
{
  fprintf(stderr,"Mix v1.2\n\n");
  fprintf(stderr,"Usage:\n\n");
  fprintf(stderr,"       %s list -> Listing of all channels\n", progname);
  fprintf(stderr,"\n");
  fprintf(stderr,"       %s recsrc [<channel> ...] -> Get/set recording source\n", progname);
  fprintf(stderr,"\n");
  fprintf(stderr,"       %s <channel> {<left>[,<right>]] ...\n",progname);
  fprintf(stderr,"       %s <channel> -> Get level\n", progname);
  fprintf(stderr,"       %s <channel> <val> -> Set level to <val> (left = right = val)\n", progname);
  fprintf(stderr,"       %s <channel> <left>,<right> -> Set unbalanced stereo channel levels\n", progname);
  fprintf(stderr,"\n");
  fprintf(stderr,"       %s file <file> -> Read commands from file\n", progname);
  fprintf(stderr,"       %s dump <file> -> Dump current settings to command file\n", progname);
  fprintf(stderr,"\n");
  fprintf(stderr,"       Levels are between 0(off) and 100(max).\n");
  exit(2);
}

void Parse(char *words[]);

void ReadFile(char *name)
{
  FILE *fid;
  char buf[301], *ptr, *words[50];
  int nw;
  if(!strcmp(name,"-")) fid = stdin;
  else if((fid = fopen(name,"r")) == 0) die("fopen");
  while(fgets(buf,300,fid) != NULL){
    buf[300] = 0;
    ptr = buf;
    while(ptr-buf<300 && (*ptr==' ' || *ptr=='\t')) ptr++;
    if(*ptr == '#') continue;
    nw = 0;
    while(1){
      words[nw++] = ptr;
      while(ptr-buf<300 && *ptr!=' ' && *ptr!='\t' && *ptr!=0
	    && *ptr!=10 && *ptr!=13)
	ptr++;
      if(*ptr==' ' || *ptr=='\t'){
	*(ptr++) = 0;
	while(ptr-buf<300 && (*ptr==' ' || *ptr=='\t')) ptr++;
      }else{
	if(ptr == words[nw-1]) nw--;
	*ptr=0;
        break;
      }
    }
    words[nw] = 0;
    Parse(words);
  }
  if(strcmp(name,"-")) fclose(fid);
}

void Parse(char *words[])
{
  if(words[0] == 0) return;
  if(words[1] == 0 && (!strcmp(words[0], "-l")
		   || !strcmp(words[0], "list")))
    ListAll();

  else if(!strcmp(words[0], "-r") || !strcmp(words[0], "recsrc")){
    if(words[1] == 0) PrintRecSrc();
    else{
      char **ptr = words+1;
      int mask = 0, d;
      while(*ptr){
	d = WhichDev(*ptr);
	if(!(recable & (1<<d)))
	  { fprintf(stderr, "%s is not a possible recording source.\n",
		    names[d]);
	    exit(1);
	  }
	mask |= (1<<d);
	ptr++;
      }
      SetRecSrc(mask);
      PrintRecSrc();
    }
  }

  else if(!strcmp(words[0], "-f") || !strcmp(words[0], "file")){
    if(words[1] == 0) Usage();
    ReadFile(words[1]);
  }

  else if(!strcmp(words[0], "-d") || !strcmp(words[0], "dump")){
    if(words[1] == 0) Usage();
    DumpFile(words[1]);
  }

  else{
    char **ptr = words;
    int dev, left, right;
    while(*ptr){
      dev = WhichDev(*ptr);
      if(*(ptr+1) && GetPair(*(ptr+1),&left,&right)){
	Set(dev,left,right);
	ptr++;
      }
      PrintDev(dev);
      ptr++;
    }
  }
}

int main(int argc, char *argv[])
{
  progname = argv[0];
  if(argc == 1) Usage(argv[0]);
  Open();
  Parse(argv+1);
  Close();

  return 0;
}
